"""JSON-LD, Linked Data Proof and Verifiable Credential constants."""

SECURITY_CONTEXT_V1_URL = "https://w3id.org/security/v1"
SECURITY_CONTEXT_V2_URL = "https://w3id.org/security/v2"
SECURITY_CONTEXT_V3_URL = "https://w3id.org/security/v3-unstable"
SECURITY_CONTEXT_URL = SECURITY_CONTEXT_V2_URL
DID_V1_CONTEXT_URL = "https://www.w3.org/ns/did/v1"
CREDENTIALS_CONTEXT_V1_URL = "https://www.w3.org/2018/credentials/v1"
CREDENTIALS_CONTEXT_V2_URL = "https://www.w3.org/ns/credentials/v2"
SECURITY_CONTEXT_BBS_URL = "https://w3id.org/security/bbs/v1"
SECURITY_CONTEXT_ED25519_2020_URL = "https://w3id.org/security/suites/ed25519-2020/v1"
SECURITY_CONTEXT_MULTIKEY_URL = "https://w3id.org/security/multikey/v1"

CREDENTIALS_ISSUER_URL = "https://www.w3.org/2018/credentials#issuer"
SECURITY_PROOF_URL = "https://w3id.org/security#proof"
SECURITY_SIGNATURE_URL = "https://w3id.org/security#signature"

VERIFIABLE_CREDENTIAL_TYPE = "VerifiableCredential"
VERIFIABLE_PRESENTATION_TYPE = "VerifiablePresentation"

EXPANDED_TYPE_CREDENTIALS_CONTEXT_V1_VC_TYPE = (
    "https://www.w3.org/2018/credentials#VerifiableCredential"
)
