TEST_LD_DOCUMENT_PROOF_BBS = {
    "@context": ["http://schema.org/", "https://w3id.org/security/v3-unstable"],
    "id": "urn:bnid:_:c14n0",
    "type": "Person",
    "email": "jane.doe@example.com",
    "firstName": "Jane",
    "jobTitle": "Professor",
    "lastName": "Does",
    "telephone": "(425) 123-4567",
    "proof": {
        "type": "BbsBlsSignatureProof2020",
        "nonce": "Qvv+50burAq9dHFIE6drJWKVcQj4G+IdwQXT/S2mHkxwbd4ZI/d4lrWQ1uiJPNToHx4=",
        "proofValue": "AAoD/5BmimoSIcEAL2qzBgzHMkRF5GiNLInT30Klkp/DKgzYdIxad+DDoKSLtu3HqUJZ44bT77EubVQclHcJl3zsI826G77z53am0jZRBCmVyTRqbjIbmlBGR+DlSPs/e7Vo35Q8hKq0wBCQBLwQ6+ArSshYVPD3fvTd6lsOpvz0fJPkDYmCXhu2w9vdXZYgsobsyAAAAHSNUZ27LZKohwy3M8OoNBk59TfX8sb/ss8QagDeQ2IOQ6Tb6tyRVeF0kkDNuYno4JAAAAACXYgOfWJ8s+gDYgb/7qXnJ8amB7hhUJxWMkFGQd1D7yopAqyp9S9lwqwY5Zi8YwxfO9JjX9g5TYBorYVooI25molA0BcvIu0IzX23j9GJ9gpJuTHCUKtizD/tgepu+H6ow/mbsIeTlpUlAlasX7NIhgAAAAJpOZ/vLm9Omn7GqKBnoxCMK+dqi4Hr9NsGNOSk2qnW9Erf74dJA+e4CRMfP4ZCbYbbx7Ex+4cckdqUH4WRby4M",
        "verificationMethod": "did:example:489398593#test",
        "proofPurpose": "assertionMethod",
        "created": "2020-10-30T04:34:10Z",
    },
}
