BBS_NESTED_VC_FULL_REVEAL_DOCUMENT_MATTR = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
        "https://w3id.org/security/bbs/v1",
    ],
    "type": ["VerifiableCredential", "UniversityDegreeCredential"],
}
