"""Constants for State Proof and LeafHash Inclusion Verification."""

ALL_ATR_KEYS = ["raw", "enc", "hash"]
LAST_SEQ_NO = "lsn"
VALUE = "value"
VAL = "val"
HASH = "hash"
LAST_UPDATE_TIME = "lut"
MARKER_CLAIM_DEF = "3"
MARKER_SCHEMA = "2"
MARKER_ATTR = "1"
MARKER_REVOC_DEF = "4"
MARKER_REVOC_REG_ENTRY_ACCUM = "6"
MARKER_REVOC_REG_ENTRY = "5"
GET_NYM = "105"
GET_ATTR = "104"
GET_SCHEMA = "107"
GET_CLAIM_DEF = "108"
GET_REVOC_REG_DEF = "115"
GET_REVOC_REG_ENTRY = "116"
GET_REVOC_REG_DELTA = "117"
NYM = "1"
ATTRIB = "100"
SCHEMA = "101"
CLAIM_DEF = "102"
REVOC_REG_DEF = "113"
REVOC_REG_ENTRY = "114"
DEST = "dest"
RESULT = "result"
DATA = "data"
SEQ_NO = "seqNo"
TXN_TIME = "txnTime"
TXN_METADATA = "txnMetadata"
TXN_TIME = "txnTime"
TXN = "txn"
NAME = "name"
VERSION = "version"
ATTR_NAMES = "attr_names"
FROM = "from"
METADATA = "metadata"
ORIGIN = "origin"
REF = "ref"
IDENTIFIER = "identifier"
CRED_DEF_ID = "credDefId"
REVOC_DEF_TYPE = "revocDefType"
REVOC_DEF_TYPE_ID = "revocRegDefId"
AUDIT_PATH = "auditPath"
ROOT_HASH = "rootHash"
STATE_PROOF = "state_proof"
STATE_PROOF_FROM = "stateProofFrom"
PROOF_NODES = "proof_nodes"
TAG = "tag"
ACCUM_TO = "accum_to"
ACCUM_FROM = "accum_from"
(NODE_TYPE_BLANK, NODE_TYPE_LEAF, NODE_TYPE_EXTENSION, NODE_TYPE_BRANCH) = tuple(range(4))
BLANK_NODE = b""
