"""Library version information."""

RECORD_TYPE_ACAPY_STORAGE_TYPE = "acapy_storage_type"
RECORD_TYPE_ACAPY_UPGRADING = "acapy_upgrading"

STORAGE_TYPE_VALUE_ANONCREDS = "askar-anoncreds"
STORAGE_TYPE_VALUE_ASKAR = "askar"
