TEST_LD_DOCUMENT_BAD_SIGNED_ED25519_2020 = {
    "@context": ["http://schema.org/", "https://w3id.org/security/bbs/v1"],
    "@type": "Person",
    "firstName": "Jane",
    "lastName": "Does",
    "jobTitle": "Professor",
    "telephone": "(425) 123-4567",
    "email": "jane.doe@example.com",
    "proof": {
        "type": "Ed25519Signature2020",
        "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
        "created": "2023-05-24T13:52:12.985755+00:00",
        "proofPurpose": "assertionMethod",
        "proofValue": "z4JqUKXM2hj49WyCFmh8aUMgsEMfNJBd3WN3xGsowBPC6s3pPdJTgBoKcWueN7amex9wVdWGZNS8p8d38sqr7ccWL",
    },
}
