TEST_LD_DOCUMENT_BAD_SIGNED_P256_2019 = {
    "@context": ["http://schema.org/", "https://w3id.org/security/bbs/v1"],
    "@type": "Person",
    "firstName": "Jane",
    "lastName": "Does",
    "jobTitle": "Professor",
    "telephone": "(425) 123-4567",
    "email": "jane.doe@example.com",
    "proof": {
        "type": "EcdsaSecp256r1Signature2019",
        "verificationMethod": "did:key:zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq#zDnaehWviigWQQD7bqF3btFquwA5w8DX2sQwkVxnAyJ7oxdjq",
        "created": "2025-01-15T00:55:37+00:00",
        "proofPurpose": "assertionMethod",
        "proofValue": "z42hBVFNcjBLGpdMcDpA9YfPagJ9Ky5Fv8PQf7KoRt3S8tYJq61shtnsSszdLxQRpospia8CrM76rKE2MRRQimQKb",
    },
}
