DIF_PRESENTATION_SUBMISSION_V1 = {
    "@context": {
        "@version": 1.1,
        "PresentationSubmission": {
            "@id": "https://identity.foundation/presentation-exchange/#presentation-submission",
            "@context": {
                "@version": 1.1,
                "presentation_submission": {
                    "@id": "https://identity.foundation/presentation-exchange/#presentation-submission",
                    "@type": "@json",
                },
            },
        },
    }
}
